#!/usr/bin/env python

import sys

from ..fragment import Fragment
from ..fivec_data import FiveCData
from ..fivec import FiveC


def run(args):
    if args.regions is None:
        regions = []
    else:
        regions = args.regions.split(',')
        if len(regions) == 1 and regions[0] == '':
            regions = []
        for i in range(len(regions)):
            try:
                regions[i] = int(regions[i])
            except:
                print sys.stderr, ("Not all arguments in -r/--regions could be converted to integers.")
                return 1
    if args.algorithm.count('binning') > 0:
        model = args.model.split(',')
        for par in model:
            if par not in ['gc', 'len', 'distance']:
                print sys.stderr, ("Not all arguments in -v/--model are valid.")
                return 1
        modelbins = args.modelbins.split(',')
        for i in range(len(modelbins)):
            try:
                modelbins[i] = int(modelbins[i])
            except:
                print sys.stderr, ("Not all arguments in -n/--modelbins could be converted to integers.")
                return 1
        if len(model) != len(modelbins):
            print sys.stderr, ("-v/--model and -n/--modelbins not equal lengths.")
            return 1
    if args.prefix is None:
        frag_fname, data_fname, project_fname = args.output
    else:
        frag_fname = "%s.frags" % args.prefix
        data_fname = "%s.fcd" % args.prefix
        project_fname = "%s.fcp" % args.prefix
    frags = Fragment(frag_fname, mode='w', silent=args.silent)
    frags.load_fragments(args.bed)
    frags.save()
    del frags
    data = FiveCData(data_fname, 'w', silent=args.silent)
    if not args.bam is None:
        data.load_data_from_bam(frag_fname, args.bam)
    else:
        data.load_data_from_counts(frag_fname, args.count)
    data.save()
    del data
    fivec = FiveC(project_fname, 'w', silent=args.silent)
    fivec.load_data(data_fname)
    fivec.filter_fragments(mininteractions=args.minint, mindistance=args.mindist, maxdistance=args.maxdist)
    fivec.find_distance_parameters()
    precorrect = False
    if args.algorithm in ['binning', 'binning-express', 'binning-probability']:
        fivec.find_binning_fragment_corrections(mindistance=args.mindist, maxdistance=args.maxdist,
                                                   regions=regions, num_bins=modelbins, model=model,
                                                   usereads=args.binreads, learning_threshold=args.threshold,
                                                   max_iterations=args.biniter, precorrect=False)
        precorrect = True
    if args.algorithm in ['probability', 'binning-probability', 'probability-binning']:
        fivec.find_probability_fragment_corrections(mindistance=args.mindist, maxdistance=args.maxdist,
                                                        regions=regions, max_iterations=args.probiter,
                                                        minchange=args.change, learningstep=args.step,
                                                        precalculate=args.precalc, precorrect=precorrect)
    elif args.algorithm in ['express', 'binning-express', 'express-binning']:
        fivec.find_express_fragment_corrections(iterations=args.expiter, mindistance=args.mindist,
                                                maxdistance=args.maxdist, remove_distance=args.nodist,
                                                usereads=args.expreads, regions=regions,
                                                precorrect=precorrect, logged=args.logged, kr=args.kr)
    if args.algorithm in ['express-binning', 'probability-binning']:
        fivec.find_binning_fragment_corrections(mindistance=args.mindist, maxdistance=args.maxdist,
                                                   regions=regions, num_bins=modelbins, model=model,
                                                   usereads=args.binreads, learning_threshold=args.threshold,
                                                   max_iterations=args.biniter, precorrect=True)
    fivec.save()



# distutils: language = c++

"""These functions provide increased speed in handling the correction
optimization functions necessary for supporting fivec analysis using HiFive.
"""

import cython
cimport numpy as np
import numpy

ctypedef np.float32_t DTYPE_t
ctypedef np.float64_t DTYPE_64_t
ctypedef np.int32_t DTYPE_int_t
ctypedef np.int64_t DTYPE_int64_t
ctypedef np.uint32_t DTYPE_uint_t
ctypedef np.int8_t DTYPE_int8_t
cdef double Inf = numpy.inf

cdef extern from "math.h":
    double exp(double x) nogil
    double log(double x) nogil
    double log10(double x) nogil
    double sqrt(double x) nogil
    double pow(double x, double x) nogil
    double abs(double x) nogil
    double round(double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil

cdef extern from "_normal.hpp":
    double stdnorm_pdf(double x) nogil
    double stdnorm_cdf(double x) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_binning_correction_adjustment(
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        np.ndarray[DTYPE_int_t, ndim=2] indices not None,
        np.ndarray[DTYPE_t, ndim=1] binning_corrections,
        np.ndarray[DTYPE_int_t, ndim=1] correction_indices,
        np.ndarray[DTYPE_int_t, ndim=1] num_bins,
        np.ndarray[DTYPE_int_t, ndim=2] frag_indices):
    cdef long long int i, frag1, frag2, bin1, bin2, index
    cdef long long int num_frags = corrections.shape[0]
    cdef long long int num_parameters = frag_indices.shape[1]
    with nogil:
        for i in range(num_frags):
            frag1 = indices[i, 0]
            frag2 = indices[i, 1]
            for j in range(num_parameters):
                bin1 = min(frag_indices[frag1, j], frag_indices[frag2, j])
                bin2 = max(frag_indices[frag1, j], frag_indices[frag2, j])
                index = (num_bins[j] - 1) * bin1 - bin1 * (bin1 - 1) / 2 + bin2 + correction_indices[j]
                corrections[i] *= binning_corrections[index]
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_gradients(
        np.ndarray[DTYPE_int_t, ndim=2] data not None,
        np.ndarray[DTYPE_t, ndim=1] log_counts_n not None,
        np.ndarray[DTYPE_t, ndim=1] log_counts not None,
        np.ndarray[DTYPE_t, ndim=1] log_counts_p not None,
        np.ndarray[DTYPE_t, ndim=1] distance_signal not None,
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        np.ndarray[DTYPE_t, ndim=1] gradients not None,
        double sigma):
    cdef long long int frag1, frag2, h, i
    cdef double value0, value, expected, z_p, z_n, pdf_p, pdf_n, cdf_p, cdf_n
    cdef long long int num_data = data.shape[0]
    cdef double sigma_2 = pow(sigma, 2.0)
    with nogil:
        for i in range(num_data):
            frag1 = data[i, 0]
            frag2 = data[i, 1]
            expected = max(0.01, corrections[frag1] + corrections[frag2] + distance_signal[i])
            z_p = (log_counts_p[i] - expected) / sigma
            cdf_p = stdnorm_cdf(z_p)
            pdf_p = stdnorm_pdf(z_p)
            z_n = (log_counts_n[i] - expected) / sigma
            cdf_n = stdnorm_cdf(z_n)
            pdf_n = stdnorm_pdf(z_n)
            value0 = (cdf_p - cdf_n) * sigma
            if value0 != 0:
                value = (pdf_p - pdf_n) / value0
                gradients[frag1] += value
                gradients[frag2] += value
            else:
                value = (expected - log_counts[i]) / sigma_2
                gradients[frag1] += value
                gradients[frag2] += value
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_prob_cost(
        np.ndarray[DTYPE_int_t, ndim=2] data not None,
        np.ndarray[DTYPE_t, ndim=1] log_counts_n not None,
        np.ndarray[DTYPE_t, ndim=1] log_counts not None,
        np.ndarray[DTYPE_t, ndim=1] log_counts_p not None,
        np.ndarray[DTYPE_t, ndim=1] distance_signal not None,
        np.ndarray[DTYPE_t, ndim=1] corrections not None,
        double sigma):
    cdef long long int frag1, frag2, i
    cdef double value, cost, expected, z_p, z_n, cdf_p, cdf_n
    cdef long long int num_data = data.shape[0]
    cdef double sigma_2 = pow(sigma, 2.0)
    with nogil:
        cost = 0.0
        for i in range(num_data):
            frag1 = data[i, 0]
            frag2 = data[i, 1]
            expected = max(0.01, corrections[frag1] + corrections[frag2] + distance_signal[i])
            z_p = (log_counts_p[i] - expected) / sigma
            cdf_p = stdnorm_cdf(z_p)
            z_n = (log_counts_n[i] - expected) / sigma
            cdf_n = stdnorm_cdf(z_n)
            value = (cdf_p - cdf_n) * sigma
            if value != 0:
                cost -= log(cdf_p - cdf_n)
            else:
                cost += 0.5 * pow(log_counts[i] - expected, 2.0) / sigma_2
    return cost


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def update_corrections(
    np.ndarray[DTYPE_int_t, ndim=1] filter not None,
    np.ndarray[DTYPE_t, ndim=1] corrections not None,
    np.ndarray[DTYPE_t, ndim=1] new_corrections not None,
    np.ndarray[DTYPE_t, ndim=1] gradients not None,
    double learning_rate ):
    cdef long long int i
    cdef long long int num_frags = filter.shape[0]
    with nogil:
        for i in range(num_frags):
            if filter[i] == 0:
                continue
            new_corrections[i] = max(-10.0, min(10.0, corrections[i] - learning_rate * gradients[i]))
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_log_fragment_means(
        np.ndarray[DTYPE_t, ndim=1] distance_means,
        np.ndarray[DTYPE_t, ndim=1] trans_means,
        np.ndarray[DTYPE_int_t, ndim=1] interactions not None,
        np.ndarray[DTYPE_64_t, ndim=1] fragment_means not None,
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=2] trans_data,
        np.ndarray[DTYPE_64_t, ndim=1] log_counts,
        np.ndarray[DTYPE_64_t, ndim=1] trans_log_counts,
        np.ndarray[DTYPE_t, ndim=1] corrections not None):
    cdef long long int i, frag1, frag2, num_data, num_trans
    cdef double cost, temp
    cdef long long int num_frags = interactions.shape[0]
    if not data is None:
        num_data = data.shape[0]
    else:
        num_data = 0
    if not trans_data is None:
        num_trans = trans_data.shape[0]
    else:
        num_trans = 0
    with nogil:
        cost = 0.0
        for i in range(num_frags):
            fragment_means[i] = 0.0
        for i in range(num_data):
            frag1 = data[i, 0]
            frag2 = data[i, 1]
            if distance_means is None:
                temp = log_counts[i] - corrections[frag1] - corrections[frag2]
            else:
                temp = log_counts[i] - distance_means[i] - corrections[frag1] - corrections[frag2]
            fragment_means[frag1] += temp
            fragment_means[frag2] += temp
        for i in range(num_trans):
            frag1 = trans_data[i, 0]
            frag2 = trans_data[i, 1]
            if trans_means is None:
                temp = trans_log_counts[i] - corrections[frag1] - corrections[frag2]
            else:
                temp = trans_log_counts[i] - trans_means[i] - corrections[frag1] - corrections[frag2]
            fragment_means[frag1] += temp
            fragment_means[frag2] += temp
        for i in range(num_frags):
            if interactions[i] == 0:
                continue
            temp = fragment_means[i] / interactions[i]
            corrections[i] += temp * 0.5
            cost += pow(temp, 2.0)
        cost = pow(cost, 0.5)
    return cost


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_fragment_means(
        np.ndarray[DTYPE_t, ndim=1] distance_means,
        np.ndarray[DTYPE_t, ndim=1] trans_means,
        np.ndarray[DTYPE_int_t, ndim=1] interactions not None,
        np.ndarray[DTYPE_64_t, ndim=1] fragment_means not None,
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=2] trans_data,
        np.ndarray[DTYPE_64_t, ndim=1] log_counts,
        np.ndarray[DTYPE_64_t, ndim=1] log_trans_counts,
        np.ndarray[DTYPE_t, ndim=1] corrections not None):
    cdef long long int i, frag1, frag2, num_data, num_trans
    cdef double cost, temp
    cdef long long int num_frags = interactions.shape[0]
    if not data is None:
        num_data = data.shape[0]
    else:
        num_data = 0
    if not trans_data is None:
        num_trans = trans_data.shape[0]
    else:
        num_trans = 0
    with nogil:
        cost = 0.0
        for i in range(num_frags):
            fragment_means[i] = 0.0
        for i in range(num_data):
            frag1 = data[i, 0]
            frag2 = data[i, 1]
            if distance_means is None:
                temp = exp(log_counts[i] - corrections[frag1] - corrections[frag2])
            else:
                temp = exp(log_counts[i] - distance_means[i] - corrections[frag1] - corrections[frag2])
            fragment_means[frag1] += temp
            fragment_means[frag2] += temp
        for i in range(num_trans):
            frag1 = trans_data[i, 0]
            frag2 = trans_data[i, 1]
            if trans_means is None:
                temp = exp(log_trans_counts[i] - corrections[frag1] - corrections[frag2])
            else:
                temp = exp(log_trans_counts[i] - trans_means[i] - corrections[frag1] - corrections[frag2])
            fragment_means[frag1] += temp
            fragment_means[frag2] += temp
        for i in range(num_frags):
            if interactions[i] == 0:
                continue
            temp = fragment_means[i] / interactions[i]
            corrections[i] += log(temp) * 0.5
            cost += pow(temp - 1.0, 2.0)
        cost = pow(cost, 0.5)
    return cost


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_v(
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=2] trans_data,
        np.ndarray[DTYPE_64_t, ndim=1] counts,
        np.ndarray[DTYPE_64_t, ndim=1] trans_counts,
        np.ndarray[DTYPE_64_t, ndim=2] corrections,
        np.ndarray[DTYPE_64_t, ndim=2] v):
    cdef long long int i, frag1, frag2, num_data, num_trans_data
    cdef double correction
    if not data is None:
        num_data = data.shape[0]
    else:
        num_data = 0
    if not trans_data is None:
        num_trans = trans_data.shape[0]
    else:
        num_trans = 0
    with nogil:
        for i in range(num_data):
            frag1 = data[i, 0]
            frag2 = data[i, 1]
            correction = corrections[frag1, 0] * corrections[frag2, 0] * counts[i]
            v[frag1, 0] += correction
            v[frag2, 0] += correction
        for i in range(num_trans):
            frag1 = trans_data[i, 0]
            frag2 = trans_data[i, 1]
            correction = corrections[frag1, 0] * corrections[frag2, 0] * trans_counts[i]
            v[frag1, 0] += correction
            v[frag2, 0] += correction
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def calculate_w(
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=2] trans_data,
        np.ndarray[DTYPE_64_t, ndim=1] counts,
        np.ndarray[DTYPE_64_t, ndim=1] trans_counts,
        np.ndarray[DTYPE_64_t, ndim=2] corrections,
        np.ndarray[DTYPE_64_t, ndim=2] p,
        np.ndarray[DTYPE_64_t, ndim=2] w):
    cdef long long int i, frag1, frag2, num_data, num_trans_data
    cdef double correction
    if not data is None:
        num_data = data.shape[0]
    else:
        num_data = 0
    if not trans_data is None:
        num_trans = trans_data.shape[0]
    else:
        num_trans = 0
    with nogil:
        for i in range(num_data):
            frag1 = data[i, 0]
            frag2 = data[i, 1]
            correction = corrections[frag1, 0] * corrections[frag2, 0] * counts[i]
            w[frag1, 0] += correction * p[frag2, 0]
            w[frag2, 0] += correction * p[frag1, 0]
        for i in range(num_trans):
            frag1 = trans_data[i, 0]
            frag2 = trans_data[i, 1]
            correction = corrections[frag1, 0] * corrections[frag2, 0] * trans_counts[i]
            w[frag1, 0] += correction * p[frag2, 0]
            w[frag2, 0] += correction * p[frag1, 0]
    return None



#!/usr/bin/env python

from ..fivec import FiveC


def run(args):
    fivec = FiveC(args.output, 'w', silent=args.silent)
    fivec.load_data(args.data)
    fivec.filter_fragments(mininteractions=args.minint, mindistance=args.mindist, maxdistance=args.maxdist)
    fivec.find_distance_parameters()
    fivec.save()



#!/usr/bin/env python

"""
This is a module contains scripts for generating plots from compact and full matrices of interaction data.

Concepts
----------

These functions take either compact, upper-triangle, or full 3d data matrices.

Data can either be arranged in compact, upper-triangle, or complete (rectangular) arrays. With HiC data, compact arrays are N x M x 2, where N is the number of fends or bins, and M is the maximum distance between fends or bins. This is useful for working with sets of short interactions. When using 5C data, the compact format is an N x M x 2 array where N is the number of forward primers and M is the number of reverse primers. Data can be raw, fend-corrected, distance-dependence removed, or enrichment values. Arrays are 3-dimensional with observed values in the first layer of d3, expected values in the second layer of d3.

API Documentation
------------------
"""

import sys
from math import exp

import h5py
import numpy
try:
    from PIL import Image
except:
    pass
try:
    from pyx import canvas, path, style, bitmap, text, trafo
except:
    pass


def plot_compact_array(data, maxscore=None, minscore=None, symmetricscaling=True, logged=True,
                       min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                       returnscale=False, diagonal_included=False, **kwargs):
    """
    Fill in and rescale bitmap from a HiC compact array.

    :param data: A three-dimensional compact array of HiC interaction data.
    :type data: numpy array
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :param diagonal_included: If true, adjust output shape as necessary.
    :type diagonal_included: bool.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting compact array..."),
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    dim = data.shape[0]
    scaled = numpy.copy(data).astype(numpy.float64)
    where = numpy.where(data[:, :, 1] > 0)
    scaled[where[0], where[1], 0] /= scaled[where[0], where[1], 1]
    if logged:
        where = numpy.where(scaled[:, :, 0] <= 0)
        scaled[where[0], where[1], 1] = 0
        where = numpy.where(scaled[:, :, 1] > 0)
        scaled[where[0], where[1], 0] = numpy.log2(scaled[where[0], where[1], 0])
    scaled[where[0], where[1], 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(scaled[where[0], where[1], 0])
    if minscore is None:
        minscore = numpy.amin(scaled[where[0], where[1], 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        scaled[where[0], where[1], 0] /= maxscore
    else:
        scaled[where[0], where[1], 0] -= minscore
        scaled[where[0], where[1], 0] /= (maxscore - minscore) * 0.5
        scaled[where[0], where[1], 0] -= 1.0
    scaled = numpy.minimum(1.0, numpy.maximum(-1.0, scaled))
    where1 = numpy.where((scaled[:, :, 1] == 1) * (scaled[:, :, 0] >= 0))
    where2 = numpy.where((scaled[:, :, 1] == 1) * (scaled[:, :, 0] < 0))
    temp0 = scaled[where1[0], where1[1], 0]
    temp1 = 1.0 - temp0
    scaled[where1[0], where1[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    temp0 = -scaled[where2[0], where2[1], 0]
    temp1 = 1.0 - temp0
    scaled[where2[0], where2[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    scaled = numpy.round(scaled).astype(numpy.uint32)
    img = numpy.empty((dim, dim), dtype=numpy.uint32)
    img[:, :] = int('ff999999', 16)
    for i in range(dim - 1 + int(diagonal_included)):
        where = numpy.where(scaled[i, :min(scaled.shape[1], dim - i - 1 + int(diagonal_included)), 1] > 0)[0]
        img[i, where + i + 1 - int(diagonal_included)] = scaled[i, where, 0]
        img[where + i + 1 - int(diagonal_included), i] = img[i, where + i + 1 - int(diagonal_included)]
    pilImage = Image.frombuffer('RGBA', (dim, dim), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_full_array(data, maxscore=None, minscore=None, symmetricscaling=True, logged=True,
                    min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                    returnscale=False, **kwargs):
    """
    Fill in and rescale bitmap from a 5C or HiC full array.

    :param data: A three-dimensional compact array of interaction data.
    :type data: numpy array
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting full array..."),
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    xdim, ydim = data.shape[:2]
    scaled = numpy.copy(data).astype(numpy.float64)
    where = numpy.where(data[:, :, 1] > 0)
    scaled[where[0], where[1], 0] /= scaled[where[0], where[1], 1]
    if logged:
        where = numpy.where(scaled[:, :, 0] <= 0)
        scaled[where[0], where[1], 1] = 0
        where = numpy.where(scaled[:, :, 1] > 0)
        scaled[where[0], where[1], 0] = numpy.log2(scaled[where[0], where[1], 0])
    scaled[where[0], where[1], 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(scaled[where[0], where[1], 0])
    if minscore is None:
        minscore = numpy.amin(scaled[where[0], where[1], 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        scaled[where[0], where[1], 0] /= maxscore
    else:
        scaled[where[0], where[1], 0] -= minscore
        scaled[where[0], where[1], 0] /= (maxscore - minscore) * 0.5
        scaled[where[0], where[1], 0] -= 1.0
    scaled = numpy.minimum(1.0, numpy.maximum(-1.0, scaled))
    img = numpy.empty((xdim, ydim), dtype=numpy.uint32)
    img.shape = (ydim, xdim)
    img[:, :] = int('ff999999', 16)
    where1 = numpy.where((scaled[:, :, 1] == 1) * (scaled[:, :, 0] >= 0))
    temp0 = scaled[where1[0], where1[1], 0]
    temp1 = 1.0 - temp0
    img[where1[1], where1[0]] = (255 * 256 ** 3 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])).astype(numpy.int64) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])).astype(numpy.int64) * 256 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])).astype(numpy.int64) * 256 ** 2)
    del where1
    where2 = numpy.where((scaled[:, :, 1] == 1) * (scaled[:, :, 0] < 0))
    temp0 = -scaled[where2[0], where2[1], 0]
    temp1 = 1.0 - temp0
    img[where2[1], where2[0]] = (255 * 256 ** 3 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])).astype(numpy.int64) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])).astype(numpy.int64) * 256 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])).astype(numpy.int64) * 256 ** 2)
    pilImage = Image.frombuffer('RGBA', (xdim, ydim), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_upper_array(data, maxscore=None, minscore=None, symmetricscaling=True, logged=True,
                     min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                     returnscale=False, diagonal_included=False, **kwargs):
    """
    Fill in and rescale bitmap from a 5C or HiC upper array.

    :param data: A two-dimensional compact array of interaction data.
    :type data: numpy array
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :param diagonal_included: If true, adjust output shape as necessary.
    :type diagonal_included: bool.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting upper array..."),
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    dim = int(0.5 + (0.25 + 2 * data.shape[0]) ** 0.5) - int(diagonal_included)
    scaled = numpy.copy(data).astype(numpy.float64)
    where = numpy.where(data[:, 1] > 0)[0]
    scaled[where, 0] /= scaled[where, 1]
    if logged:
        where = numpy.where(scaled[:, 0] <= 0)[0]
        scaled[where, 1] = 0
        where = numpy.where(scaled[:, 1] > 0)[0]
        scaled[where, 0] = numpy.log2(scaled[where, 0])
    scaled[where, 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(scaled[where, 0])
    if minscore is None:
        minscore = numpy.amin(scaled[where, 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        scaled[where, 0] /= maxscore
    else:
        scaled[where, 0] -= minscore
        scaled[where, 0] /= (maxscore - minscore) * 0.5
        scaled[where, 0] -= 1.0
    scaled = numpy.minimum(1.0, numpy.maximum(-1.0, scaled))
    where1 = numpy.where((scaled[:, 1] == 1) * (scaled[:, 0] >= 0))
    where2 = numpy.where((scaled[:, 1] == 1) * (scaled[:, 0] < 0))
    temp0 = scaled[where1[0], 0]
    temp1 = 1.0 - temp0
    scaled[where1[0], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    temp0 = -scaled[where2[0], 0]
    temp1 = 1.0 - temp0
    scaled[where2[0], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    scaled = numpy.round(scaled).astype(numpy.uint32)
    img = numpy.empty((dim, dim), dtype=numpy.uint32)
    img[:, :] = int('ff999999', 16)
    for i in range(dim - 1 + int(diagonal_included)):
        index = i * dim - i * (i + 1 - 2 * int(diagonal_included)) / 2
        where = numpy.where(scaled[index:(index + dim - i - 1 + int(diagonal_included)), 1] > 0)[0]
        img[i, where + i + 1 - int(diagonal_included)] = scaled[where + index, 0]
        img[where + i + 1 - int(diagonal_included), i] = img[i, where + i + 1 - int(diagonal_included)]
    pilImage = Image.frombuffer('RGBA', (dim, dim), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_hic_heatmap(filename, maxscore=None, minscore=None, symmetricscaling=True, logged=True, chroms=[],
                     min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                     returnscale=False, format='hdf5', **kwargs):
    """
    Fill in and rescale bitmap from a HiC heatmap h5dict file.

    :param filename: File name of heatmap h5dict containing binned data arrays.
    :type filename: str.
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param chroms: A list of chromosome names to include in the plot. If left empty, all chromosomes present in the heatmap file will be plotted.
    :type chroms: list
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :param format: Format of the heatmap.
    :type format: str.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.\n")
        return None
    if format not in ['hdf5', 'npz']:
        if not silent:
            print >> sys.stderr, ("The format is not recognized or supported.\n")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting heatmap dict..."),
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    if format == 'hdf5':
        input = h5py.File(filename, 'r')
    else:
        input = numpy.load(filename)
    if len(chroms) == 0:
        chroms = list(input['chromosomes'][:])
    starts = [0]
    sizes = []
    for chrom in chroms:
        sizes.append(input['%s.positions' % chrom].shape[0])
        starts.append(starts[-1] + sizes[-1] + 1)
    data = numpy.zeros((starts[-1] - 1, starts[-1] - 1, 2), dtype=numpy.float64)
    for i in range(len(chroms)):
        if '%s.counts' % chroms[i] in input and '%s.expected' % chroms[i] in input:
            n = input['%s.positions' % chroms[i]].shape[0]
            if n * (n - 1) / 2 == input['%s.counts' % chroms[i]].shape[0]:
                indices = numpy.triu_indices(sizes[i], 1)
            else:
                indices = numpy.triu_indices(sizes[i], 0)
            data[indices[0] + starts[i], indices[1] + starts[i], 0] = input['%s.counts' % chroms[i]][:]
            data[indices[0] + starts[i], indices[1] + starts[i], 1] = input['%s.expected' % chroms[i]][:]
    for i in range(len(chroms) - 1):
        for j in range(i + 1, len(chroms)):
            if '%s_by_%s.counts' % (chroms[i], chroms[j]) in input.keys():
                data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), 0] = (
                    input['%s_by_%s.counts' % (chroms[i], chroms[j])][:, :])
                data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), 1] = (
                    input['%s_by_%s.expected' % (chroms[i], chroms[j])][:, :])
                data[starts[j]:(starts[j + 1] - 1), starts[i]:(starts[i + 1] - 1), :] = numpy.transpose(
                    data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), :], axes=[1, 0, 2])
            elif '%s_by_%s.counts' % (chroms[j], chroms[i]) in input.keys():
                data[starts[j]:(starts[j + 1] - 1), starts[i]:(starts[i + 1] - 1), 0] = (
                    input['%s_by_%s.counts' % (chroms[j], chroms[i])][:, :])
                data[starts[j]:(starts[j + 1] - 1), starts[i]:(starts[i + 1] - 1), 1] = (
                    input['%s_by_%s.expected' % (chroms[j], chroms[i])][:, :])
                data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), :] = numpy.transpose(
                    data[starts[j]:(starts[j + 1] - 1), starts[i]:(starts[i + 1] - 1), :], axes=[1, 0, 2])
    where = numpy.where(data[:, :, 1] > 0)
    data[where[0], where[1], 0] /= data[where[0], where[1], 1]
    if logged:
        where = numpy.where(data[:, :, 0] <= 0)
        data[where[0], where[1], 1] = 0
        where = numpy.where(data[:, :, 1] > 0)
        data[where[0], where[1], 0] = numpy.log2(data[where[0], where[1], 0])
    data[where[0], where[1], 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(data[where[0], where[1], 0])
    if minscore is None:
        minscore = numpy.amin(data[where[0], where[1], 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        data[where[0], where[1], 0] /= maxscore
    else:
        data[where[0], where[1], 0] -= minscore
        data[where[0], where[1], 0] /= (maxscore - minscore) * 0.5
        data[where[0], where[1], 0] -= 1.0
    data = numpy.minimum(1.0, numpy.maximum(-1.0, data))
    where1 = numpy.where((data[:, :, 1] == 1) * (data[:, :, 0] >= 0))
    where2 = numpy.where((data[:, :, 1] == 1) * (data[:, :, 0] < 0))
    temp0 = data[where1[0], where1[1], 0]
    temp1 = 1.0 - temp0
    data[where1[0], where1[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    temp0 = -data[where2[0], where2[1], 0]
    temp1 = 1.0 - temp0
    data[where2[0], where2[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    data = numpy.round(data).astype(numpy.uint32)
    img = numpy.empty((data.shape[0], data.shape[0]), dtype=numpy.uint32)
    img[:, :] = int('ff999999', 16)
    where = numpy.where(data[:, :, 1] > 0)
    img[where] = data[where[0], where[1], 0]
    img[where[1], where[0]] = img[where]
    pilImage = Image.frombuffer('RGBA', (data.shape[0], data.shape[0]), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_fivec_heatmap(filename, maxscore=None, minscore=None, symmetricscaling=True, logged=True,
                       regions=[], min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                       returnscale=False, **kwargs):
    """
    Fill in and rescale bitmap in a full format from a 5C heatmap h5dict file.

    :param filename: Location of a heatmap h5dict containing 5C data arrays.
    :type data: str.
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param regions: If specified, only the indicated regions are plotted. Otherwise all regions present in the h5dict are plotted.
    :type regions: list
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting heatmap dict..."),
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    input = h5py.File(filename, 'r')
    if len(regions) == 0:
        regions = []
        for key in input.keys():
            if key.count('positions') > 0 and key.count('reverse') == 0:
                regions.append(int(key.split('.')[0]))
        regions.sort()
    else:
        for i in range(len(regions))[::-1]:
            if ("%i.positions" % regions[i] not in input.keys() and
                "%i.forward_positions" % regions[i] not in input.keys()):
                if not silent:
                    print >> sys.stderr, ("Region %i not in heatmap, skipping\n"),
                del regions[i]
    if len(regions) == 0:
        if not silent:
            print >> sys.stderr, ("No valid data found.\n"),
        return None
    if "%i.positions" % regions[0] in input.keys():
        starts = [0]
    else:
        starts = [[0, 0]]
    sizes = []
    for region in regions:
        if "%i.positions" % region in input:
            sizes.append(input['%i.positions' % region].shape[0])
            starts.append(starts[-1] + sizes[-1] + 1)
        else:
            sizes.append([input['%i.forward_positions' % region].shape[0],
                          input['%i.reverse_positions' % region].shape[0]])
            starts.append([starts[-1][0] + sizes[-1][0] + 1, starts[-1][1] + sizes[-1][1] + 1])
    sizes = numpy.array(sizes, dtype=numpy.int32)
    starts = numpy.array(starts, dtype=numpy.int32)
    if len(sizes.shape) == 1:
        data = numpy.zeros((starts[-1] - 1, starts[-1] - 1, 2), dtype=numpy.float64)
    else:
        data = numpy.zeros((starts[-1, 0] - 1, starts[-1, 1] - 1, 2), dtype=numpy.float64)
    for i in range(len(regions)):
        if '%i.counts' % regions[i] in input and '%i.expected' % regions[i] in input:
            if len(sizes.shape) == 1:
                data[starts[i]:(starts[i + 1] - 1),
                     starts[i]:(starts[i + 1]  - 1), 0] = input['%i.counts' % regions[i]][...]
                data[starts[i]:(starts[i + 1] - 1),
                     starts[i]:(starts[i + 1] - 1), 1] = input['%i.expected' % regions[i]][...]
            else:
                data[starts[i, 0]:(starts[i + 1, 0] - 1),
                     starts[i, 1]:(starts[i + 1, 1]  - 1), 0] = input['%i.counts' % regions[i]][...]
                data[starts[i, 0]:(starts[i + 1, 0] - 1),
                     starts[i, 1]:(starts[i + 1, 1] - 1), 1] = input['%i.expected' % regions[i]][...]
    for i in range(len(regions) - 1):
        for j in range(i + 1, len(regions)):
            if '%i_by_%i.counts' % (regions[i], regions[j]) in input.keys():
                if len(sizes.shape) == 1:
                    data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), 0] = (
                        input['%i_by_%i.counts' % (regions[i], regions[j])][:, :])
                    data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), 1] = (
                        input['%i_by_%i.expected' % (regions[i], regions[j])][:, :])
                    data[starts[j]:(starts[j + 1] - 1), starts[i]:(starts[i + 1] - 1), :] = numpy.transpose(
                        data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), :], axes=[1, 0, 2])
                else:
                    data[starts[i, 0]:(starts[i + 1, 0] - 1), starts[j, 1]:(starts[j + 1, 1] - 1), 0] = (
                        input['%i_by_%i.counts' % (regions[i], regions[j])][:, :])
                    data[starts[i, 0]:(starts[i + 1, 0] - 1), starts[j, 1]:(starts[j + 1, 1] - 1), 1] = (
                        input['%i_by_%i.expected' % (regions[i], regions[j])][:, :])
            if '%i_by_%i.counts' % (regions[j], regions[i]) in input.keys():
                if len(sizes.shape) == 1:
                    data[starts[j]:(starts[j + 1] - 1), starts[i]:(starts[i + 1] - 1), 0] = (
                        input['%i_by_%i.counts' % (regions[j], regions[i])][:, :])
                    data[starts[j]:(starts[j + 1] - 1), starts[i]:(starts[i + 1] - 1), 1] = (
                        input['%i_by_%i.expected' % (regions[j], regions[i])][:, :])
                    data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), :] = numpy.transpose(
                        data[starts[j]:(starts[j + 1] - 1), starts[i]:(starts[i + 1] - 1), :], axes=[1, 0, 2])
                else:
                    data[starts[j, 0]:(starts[j + 1, 0] - 1), starts[i, 1]:(starts[i + 1, 1] - 1), 0] = (
                        input['%i_by_%i.counts' % (regions[j], regions[i])][:, :])
                    data[starts[j, 0]:(starts[j + 1, 0] - 1), starts[i, 1]:(starts[i + 1, 1] - 1), 1] = (
                        input['%i_by_%i.expected' % (regions[j], regions[i])][:, :])
    where = numpy.where(data[:, :, 1] > 0)
    data[where[0], where[1], 0] /= data[where[0], where[1], 1]
    if logged:
        where = numpy.where(data[:, :, 0] <= 0)
        data[where[0], where[1], 1] = 0
        where = numpy.where(data[:, :, 1] > 0)
        data[where[0], where[1], 0] = numpy.log2(data[where[0], where[1], 0])
    data[where[0], where[1], 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(data[where[0], where[1], 0])
    if minscore is None:
        minscore = numpy.amin(data[where[0], where[1], 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        data[where[0], where[1], 0] /= maxscore
    else:
        data[where[0], where[1], 0] -= minscore
        data[where[0], where[1], 0] /= (maxscore - minscore) * 0.5
        data[where[0], where[1], 0] -= 1.0
    data = numpy.minimum(1.0, numpy.maximum(-1.0, data))
    where1 = numpy.where((data[:, :, 1] == 1) * (data[:, :, 0] >= 0))
    where2 = numpy.where((data[:, :, 1] == 1) * (data[:, :, 0] < 0))
    temp0 = data[where1[0], where1[1], 0]
    temp1 = 1.0 - temp0
    data[where1[0], where1[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    temp0 = -data[where2[0], where2[1], 0]
    temp1 = 1.0 - temp0
    data[where2[0], where2[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    data = numpy.round(data).astype(numpy.uint32)
    img = numpy.empty((data.shape[0], data.shape[1]), dtype=numpy.uint32)
    img.shape = (img.shape[1], img.shape[0])
    img[:, :] = int('ff999999', 16)
    where = numpy.where(data[:, :, 1] > 0)
    img[where[1], where[0]] = data[where[0], where[1], 0]
    black = int('ff000000', 16)
    if len(sizes.shape) == 1:
        for i in range(1, starts.shape[0] - 1):
            img[starts[i] - 1, :] = black
            img[:, starts[i] - 1] = black
    else:
        for i in range(1, starts.shape[0] - 1):
            img[starts[i, 1] - 1, :] = black
            img[:, starts[i, 0] - 1] = black
    pilImage = Image.frombuffer('RGBA', (data.shape[0], data.shape[1]), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_fivec_dict(array_dict, maxscore=None, minscore=None, symmetricscaling=True, logged=True,
                       regions=[], min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                       returnscale=False, **kwargs):
    """
    Fill in and bitmap ifrom a 5C heatmap dictionary.

    :param array_dict: Dictionary containing 5C data arrays and either 1 or two 2D mapping arrays (full or compact, respectively) with tuples are region name or pairs as keys.
    :type data: dict.
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param regions: If specified, only the indicated regions are plotted. Otherwise all regions present in the h5dict are plotted.
    :type regions: list
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting heatmap dict..."),
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    if len(regions) == 0:
        regions = []
        for key in array_dict.keys():
            regions.append(key[0])
            if len(key) > 1:
                regions.append(key[1])
        regions = list(set(regions))
        regions.sort()
    else:
        for i in range(len(regions))[::-1]:
            if (i,) not in array_dict:
                if not silent:
                    print >> sys.stderr, ("Region %i not in heatmap, skipping\n"),
                del regions[i]
    if len(regions) == 0:
        if not silent:
            print >> sys.stderr, ("No valid data found.\n"),
        return None
    if len(array_dict[(regions[0],)]) == 2:
        starts = [0]
        arraytype = 'full'
    else:
        starts = [[0, 0]]
        arraytype = 'compact'
    sizes = []
    for region in regions:
        if arraytype == 'full':
            sizes.append(array_dict[(region,)][1].shape[0])
            starts.append(starts[-1] + sizes[-1] + 1)
        else:
            sizes.append([array_dict[(region,)][1].shape[0],
                          array_dict[(region,)][2].shape[0]])
            starts.append([starts[-1][0] + sizes[-1][0] + 1, starts[-1][1] + sizes[-1][1] + 1])
    sizes = numpy.array(sizes, dtype=numpy.int32)
    starts = numpy.array(starts, dtype=numpy.int32)
    if len(sizes.shape) == 1:
        data = numpy.zeros((starts[-1] - 1, starts[-1] - 1, 2), dtype=numpy.float64)
    else:
        data = numpy.zeros((starts[-1, 0] - 1, starts[-1, 1] - 1, 2), dtype=numpy.float64)
    for i in range(len(regions)):
        if arraytype == 'full':
            data[starts[i]:(starts[i + 1] - 1),
                 starts[i]:(starts[i + 1]  - 1), :] = array_dict[(regions[i],)][0]
        else:
            data[starts[i, 0]:(starts[i + 1, 0] - 1),
                 starts[i, 1]:(starts[i + 1, 1]  - 1), :] = array_dict[(regions[i],)][0]
        for j in range(i + 1, len(regions)):
            key = (regions[i], regions[j])
            if key in array_dict:
                if arraytype == 'full':
                    data[starts[i]:(starts[i + 1] - 1), starts[j]:(starts[j + 1] - 1), :] = array_dict[key]
                    data[starts[j]:(starts[j + 1] - 1),
                         starts[i]:(starts[i + 1] - 1), :] = array_dict[key].transpose(1, 0, 2)
                else:
                    data[starts[i, 0]:(starts[i + 1, 0] - 1), starts[j, 1]:(starts[j + 1, 1] - 1), :] = array_dict[key]
                    data[starts[j, 0]:(starts[j + 1, 0] - 1),
                         starts[i, 1]:(starts[i + 1, 1] - 1), :] = array_dict[(regions[j], regions[i])]
    where = numpy.where(data[:, :, 1] > 0)
    data[where[0], where[1], 0] /= data[where[0], where[1], 1]
    if logged:
        where = numpy.where(data[:, :, 0] <= 0)
        data[where[0], where[1], 1] = 0
        where = numpy.where(data[:, :, 1] > 0)
        data[where[0], where[1], 0] = numpy.log2(data[where[0], where[1], 0])
    data[where[0], where[1], 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(data[where[0], where[1], 0])
    if minscore is None:
        minscore = numpy.amin(data[where[0], where[1], 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        data[where[0], where[1], 0] /= maxscore
    else:
        data[where[0], where[1], 0] -= minscore
        data[where[0], where[1], 0] /= (maxscore - minscore) * 0.5
        data[where[0], where[1], 0] -= 1.0
    data = numpy.minimum(1.0, numpy.maximum(-1.0, data))
    where1 = numpy.where((data[:, :, 1] == 1) * (data[:, :, 0] >= 0))
    where2 = numpy.where((data[:, :, 1] == 1) * (data[:, :, 0] < 0))
    temp0 = data[where1[0], where1[1], 0]
    temp1 = 1.0 - temp0
    data[where1[0], where1[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    temp0 = -data[where2[0], where2[1], 0]
    temp1 = 1.0 - temp0
    data[where2[0], where2[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    data = numpy.round(data).astype(numpy.uint32)
    img = numpy.empty((data.shape[0], data.shape[1]), dtype=numpy.uint32)
    img.shape = (img.shape[1], img.shape[0])
    img[:, :] = int('ff999999', 16)
    where = numpy.where(data[:, :, 1] > 0)
    img[where[1], where[0]] = data[where[0], where[1], 0]
    black = int('ff000000', 16)
    if len(sizes.shape) == 1:
        for i in range(1, starts.shape[0] - 1):
            img[starts[i] - 1, :] = black
            img[:, starts[i] - 1] = black
    else:
        for i in range(1, starts.shape[0] - 1):
            img[starts[i, 1] - 1, :] = black
            img[:, starts[i, 0] - 1] = black
    pilImage = Image.frombuffer('RGBA', (data.shape[0], data.shape[1]), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_fivec_compact_heatmap_dict(filename, maxscore=None, minscore=None, symmetricscaling=True, logged=True,
                                    regions=[], min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                                    returnscale=False, **kwargs):
    """
    Fill in and rescale bitmap in a compact format from a 5C heatmap h5dict file.

    This plots the data in a 5C compact format such that the rows correspond to positive-strand primers and columns correspond to negative-strand primers.

    :param filename: Location of a heatmap h5dict containing 5C data arrays.
    :type data: str.
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param regions: If specified, only the indicated regions are plotted. Otherwise all regions present in the h5dict are plotted.
    :type regions: list
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting heatmap dict..."),
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    input = h5py.File(filename, 'r')
    if len(regions) == 0:
        regions = []
        for key in input.keys():
            if key.count('forward_fragments') > 0:
                regions.append(int(key.split('.')[0]))
        regions.sort()
    xstarts = [0]
    ystarts = [0]
    xsizes = []
    ysizes = []
    for region in regions:
        xsizes.append(input['%i.forward_fragments' % region].shape[0])
        xstarts.append(xstarts[-1] + xsizes[-1] + 1)
        ysizes.append(input['%i.reverse_fragments' % region].shape[0])
        ystarts.append(ystarts[-1] + ysizes[-1] + 1)
    data = numpy.zeros((xstarts[-1] - 1, ystarts[-1] - 1, 2), dtype=numpy.float64)
    for i in range(len(regions)):
        if '%i.counts' % regions[i] in input and '%i.expected' % regions[i] in input:
            data[xstarts[i]:(xstarts[i + 1] - 1), ystarts[i]:(ystarts[i + 1]  - 1), 0] = (
                    input['%i.counts' % regions[i]][...])
            data[xstarts[i]:(xstarts[i + 1] - 1), ystarts[i]:(ystarts[i + 1]  - 1), 1] = (
                    input['%i.expected' % regions[i]][...])
    for i in range(len(regions) - 1):
        for j in range(i + 1, len(regions)):
            if ('%i_by_%i.counts' % (regions[i], regions[j]) in input and
                '%i_by_%i.expected' % (regions[i], regions[j]) in input):
                data[xstarts[i]:(xstarts[i + 1] - 1), ystarts[j]:(ystarts[j + 1] - 1), 0] = (
                    input['%i_by_%i.counts' % (regions[i], regions[j])][:, :])
                data[xstarts[i]:(xstarts[i + 1] - 1), ystarts[j]:(ystarts[j + 1] - 1), 1] = (
                    input['%i_by_%i.expected' % (regions[i], regions[j])][:, :])
            if ('%i_by_%i.counts' % (regions[j], regions[i]) in input and
                '%i_by_%i.expected' % (regions[j], regions[i]) in input):
                data[xstarts[j]:(xstarts[j + 1] - 1), ystarts[i]:(ystarts[i + 1] - 1), 0] = (
                    input['%i_by_%i.counts' % (regions[j], regions[i])][:, :])
                data[xstarts[j]:(xstarts[j + 1] - 1), ystarts[i]:(ystarts[i + 1] - 1), 1] = (
                    input['%i_by_%i.expected' % (regions[j], regions[i])][:, :])
    where = numpy.where(data[:, :, 1] > 0)
    data[where[0], where[1], 0] /= data[where[0], where[1], 1]
    if logged:
        where = numpy.where(data[:, :, 0] <= 0)
        data[where[0], where[1], 1] = 0
        where = numpy.where(data[:, :, 1] > 0)
        data[where[0], where[1], 0] = numpy.log2(data[where[0], where[1], 0])
    data[where[0], where[1], 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(data[where[0], where[1], 0])
    if minscore is None:
        minscore = numpy.amin(data[where[0], where[1], 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        data[where[0], where[1], 0] /= maxscore
    else:
        data[where[0], where[1], 0] -= minscore
        data[where[0], where[1], 0] /= (maxscore - minscore) * 0.5
        data[where[0], where[1], 0] -= 1.0
    data = numpy.minimum(1.0, numpy.maximum(-1.0, data))
    where1 = numpy.where((data[:, :, 1] == 1) * (data[:, :, 0] >= 0))
    where2 = numpy.where((data[:, :, 1] == 1) * (data[:, :, 0] < 0))
    temp0 = data[where1[0], where1[1], 0]
    temp1 = 1.0 - temp0
    data[where1[0], where1[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    temp0 = -data[where2[0], where2[1], 0]
    temp1 = 1.0 - temp0
    data[where2[0], where2[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    data = numpy.round(data).astype(numpy.uint32)
    img = numpy.empty((data.shape[0], data.shape[1]), dtype=numpy.uint32)
    img.shape = (data.shape[1], data.shape[0])
    img[:, :] = int('ff999999', 16)
    where = numpy.where(data[:, :, 1] > 0)
    img[where[1], where[0]] = data[where[0], where[1], 0]
    black = int('ff000000', 16)
    for i in range(1, len(xstarts) - 1):
        img[:, xstarts[i] - 1] = black
    for i in range(1, len(ystarts) - 1):
        img[ystarts[i] - 1, :] = black
    pilImage = Image.frombuffer('RGBA', (data.shape[0], data.shape[1]), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_diagonal_from_upper_array(data, maxscore=None, minscore=None, symmetricscaling=True, logged=True,
                                   min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                                   returnscale=False, diagonal_included=False, **kwargs):
    """
    Fill in and rescale bitmap from a HiC upper array, plotting only the upper triangle rotated 45 degrees counter-clockwise.

    :param data: A three-dimensional upper array of HiC interaction data.
    :type data: numpy array
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :param diagonal_included: If true, adjust output shape as necessary.
    :type diagonal_included: bool.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting rotated compact array..."),
    diag = int(diagonal_included)
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    dim = int(0.5 + (0.25 + data.shape[0] * 2) ** 0.5) - diag
    rotated = numpy.zeros((dim * 2 - 1 + diag * 2, dim + diag, 2), dtype=numpy.float64)
    xdim, ydim = rotated.shape[:2]
    for i in range(dim - 1 + diag):
        start = i * dim - (i * (i + 1 - 2 * diag) / 2)
        stop = start + dim - i - 1 + diag
        span = stop - start
        rotated[numpy.arange(i * 2 + 1, i * 2 + span + 1), ydim - numpy.arange(span) - 1, :] += data[start:stop, :]
        rotated[numpy.arange(i * 2, i * 2 + span), ydim - numpy.arange(span) - 1, :] += data[start:stop, :] * 0.25
        rotated[numpy.arange(i * 2 + 2, i * 2 + span + 2), ydim - numpy.arange(span) - 1, :] += data[start:stop, :] * 0.25
        rotated[numpy.arange(i * 2 + 1, i * 2 + span + 1), ydim - numpy.arange(span) - 2, :] += data[start:stop, :] * 0.25
        rotated[numpy.arange(i * 2 + 2, i * 2 + span + 1), ydim - numpy.arange(span - 1) - 1, :] += data[(start + 1):stop, :] * 0.25
    where = numpy.where(rotated[:, :, 1] > 0)
    rotated[where[0], where[1], 0] /= rotated[where[0], where[1], 1]
    if logged:
        where = numpy.where(rotated[:, :, 0] <= 0)
        rotated[where[0], where[1], 1] = 0
        where = numpy.where(rotated[:, :, 1] > 0)
        rotated[where[0], where[1], 0] = numpy.log2(rotated[where[0], where[1], 0])
    rotated[where[0], where[1], 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(rotated[where[0], where[1], 0])
    if minscore is None:
        minscore = numpy.amin(rotated[where[0], where[1], 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        rotated[where[0], where[1], 0] /= maxscore
    else:
        rotated[where[0], where[1], 0] -= minscore
        rotated[where[0], where[1], 0] /= (maxscore - minscore) * 0.5
        rotated[where[0], where[1], 0] -= 1.0
    rotated = numpy.minimum(1.0, numpy.maximum(-1.0, rotated))
    where1 = numpy.where((rotated[:, :, 1] == 1) * (rotated[:, :, 0] >= 0))
    where2 = numpy.where((rotated[:, :, 1] == 1) * (rotated[:, :, 0] < 0))
    temp0 = rotated[where1[0], where1[1], 0]
    temp1 = 1.0 - temp0
    rotated[where1[0], where1[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    temp0 = -rotated[where2[0], where2[1], 0]
    temp1 = 1.0 - temp0
    rotated[where2[0], where2[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    rotated = numpy.round(rotated).astype(numpy.uint32)
    img = numpy.empty((xdim, ydim), dtype=numpy.uint32)
    img.shape = (ydim, xdim)
    img[:, :] = int('ff999999', 16)
    where = numpy.where(rotated[:, :, 1] > 0)
    img[where[1], where[0]] = rotated[where[0], where[1], 0]
    pilImage = Image.frombuffer('RGBA', (xdim, ydim), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_diagonal_from_compact_array(data, maxscore=None, minscore=None, symmetricscaling=True, logged=True,
                                     min_color="0000ff", mid_color="ffffff", max_color="ff0000",
                                     returnscale=False, diagonal_included=False, **kwargs):
    """
    Fill in and rescale bitmap from a HiC compact array, plotting only the upper triangle rotated 45 degrees counter-clockwise.

    :param data: A three-dimensional compact array of HiC interaction data.
    :type data: numpy array
    :param maxscore: A ceiling value to cutoff scores at for plot color.
    :type maxscore: float
    :param minscore: A floor value to cutoff scores at for plot color.
    :type minscore: float
    :param symmetricscaling: Indicates whether to recenter data for scaling or maintain scores about zero.
    :type symmetricscaling: bool.
    :param logged: Indicates whether to use log2 values of scores for color values.
    :type logged: bool.
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param returnscale: Indicates whether to return a list containing the bitmap, minimum score, and maximum score, or just the bitmap.
    :type returnscale: bool.
    :param diagonal_included: If true, adjust output shape as necessary.
    :type diagonal_included: bool.
    :returns: :mod:`PIL` bitmap object and if requested, minimum and maximum scores.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL module must be installed to use this function.")
        return None
    if not silent:
        print >> sys.stderr, ("Plotting rotated compact array..."),
    diag = int(diagonal_included)
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    rotated = numpy.zeros((data.shape[0] * 2 - 1 + diag * 2, data.shape[1] + 1, 2), dtype=numpy.float64)
    xdim, ydim = rotated.shape[:2]
    for i in range(data.shape[1]):
        rotated[(i + 1):(2 * data.shape[0] - i - 1 + diag * 2):2, ydim - i - 1, :] += (
            data[:(data.shape[0] - i - 1 + diag), i, :])
        rotated[i:(2 * data.shape[0] - i - 2 + diag * 2):2, ydim - i - 1, :] += (
            data[:(data.shape[0] - i - 1 + diag), i, :]) * 0.25
        rotated[(i + 2):(2 * data.shape[0] - i + diag * 2):2, ydim - i - 1, :] += (
            data[:(data.shape[0] - i - 1 + diag), i, :]) * 0.25
        rotated[(i + 1):(2 * data.shape[0] - i - 1 + diag * 2):2, ydim - i - 2, :] += (
            data[:(data.shape[0] - i - 1 + diag), i, :]) * 0.25
        if i > 0:
            rotated[(i + 1):(2 * data.shape[0] - i - 1 + diag * 2):2, ydim - i, :] += (
                data[:(data.shape[0] - i - 1 + diag), i, :]) * 0.25
    where = numpy.where(rotated[:, :, 1] > 0)
    rotated[where[0], where[1], 0] /= rotated[where[0], where[1], 1]
    if logged:
        where = numpy.where(rotated[:, :, 0] <= 0)
        rotated[where[0], where[1], 1] = 0
        where = numpy.where(rotated[:, :, 1] > 0)
        rotated[where[0], where[1], 0] = numpy.log2(rotated[where[0], where[1], 0])
    rotated[where[0], where[1], 1] = 1
    if maxscore is None:
        maxscore = numpy.amax(rotated[where[0], where[1], 0])
    if minscore is None:
        minscore = numpy.amin(rotated[where[0], where[1], 0])
    if symmetricscaling:
        maxscore = max(abs(maxscore), abs(minscore))
        minscore = -maxscore
        rotated[where[0], where[1], 0] /= maxscore
    else:
        rotated[where[0], where[1], 0] -= minscore
        rotated[where[0], where[1], 0] /= (maxscore - minscore) * 0.5
        rotated[where[0], where[1], 0] -= 1.0
    rotated = numpy.minimum(1.0, numpy.maximum(-1.0, rotated))
    where1 = numpy.where((rotated[:, :, 1] == 1) * (rotated[:, :, 0] >= 0))
    where2 = numpy.where((rotated[:, :, 1] == 1) * (rotated[:, :, 0] < 0))
    temp0 = rotated[where1[0], where1[1], 0]
    temp1 = 1.0 - temp0
    rotated[where1[0], where1[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    temp0 = -rotated[where2[0], where2[1], 0]
    temp1 = 1.0 - temp0
    rotated[where2[0], where2[1], 0] = (255.0 * 256.0 ** 3.0 +
        numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
        numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256.0 +
        numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256.0 ** 2.0)
    rotated = numpy.round(rotated).astype(numpy.uint32)
    img = numpy.empty((xdim, ydim), dtype=numpy.uint32)
    img.shape = (ydim, xdim)
    img[:, :] = int('ff999999', 16)
    where = numpy.where(rotated[:, :, 1] > 0)
    img[where[1], where[0]] = rotated[where[0], where[1], 0]
    pilImage = Image.frombuffer('RGBA', (xdim, ydim), img, 'raw', 'RGBA', 0, 1)
    if not silent:
        print >> sys.stderr, ("Done\n"),
    if returnscale:
        return [pilImage, minscore, maxscore]
    else:
        return pilImage


def plot_key(min_score, max_score, height, width, labelformat='%0.2f', orientation='left', num_ticks=5,
             min_color="0000ff", mid_color="ffffff", max_color="ff0000", labelattr=None, log_display=True, **kwargs):
    """
    Create a key including color gradient and labels indicating associated values, returning a :mod:`pyx` canvas.

    :param min_score: The minimum value of the key scale.
    :type min_score: float
    :param max_score: The maximum value of the key scale.
    :type max_score: float
    :param height: The height of the gradient bar in whatever units :mod:`pyx` is using.
    :type height: float
    :param width: The width of the gradient bar in whatever units :mod:`pyx` is using.
    :type width: float
    :param labelformat: A string denoting the format for displaying number labels using the string formatting style from Python <= 2.6.
    :type labelformat: str.
    :param orientation: Indicates where labels are placed relative to gradient bar. This parameter will accept 'left', 'right', 'top', and 'bottom'.
    :type type: str.
    :param num_ticks: Indicates how many evenly-spaced tick marks and associated labels to insert. This can be zero for no labels or greater than one. Labels are inserted at the minimum and maximum values first with remaining ticks occuring evenly distributed between the extremes.
    :type num_ticks: int.
    :param min_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This variable is used to create a color gradient for plotting along with max_color and mid_color.
    :type min_color: str.
    :param mid_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the minimum plot value. This can be set to None to create a gradient ranging from min_color to max_color or to a hex color to create a divergent gradient.
    :type mid_color: str.
    :param max_color: This is a hex color code ("rrggbb" where each pair ranges from 00-ff) specifying the color associated with the maximum plot value. This variable is used to create a color gradient for plotting along with min_color and mid_color.
    :type max_color: str.
    :param labelattr: A list of pyx attributes to be passed to the text function.
    :type labelattr: str.
    :param log_display: If True, min_score and max_score are taken to be logged values and so labels are evenly spaced in log space but converted to normal space for display.
    :type log_display: bool.
    :returns: :mod:`Pxy` canvas object.
    """
    if 'silent' in kwargs and kwargs['silent']:
        silent = True
    else:
        silent = False
    if 'PIL' not in sys.modules.keys() or 'pyx' not in sys.modules.keys():
        if not silent:
            print >> sys.stderr, ("The PIL and pyx modules must be installed to use this function.")
        return None
    height = float(height)
    width = float(width)
    min_score = float(min_score)
    max_score = float(max_score)
    c = canvas.canvas()
    if orientation in ['right', 'left']:
        img = numpy.zeros( (int(round(511.0 * width / height)), 511), dtype=numpy.uint32 )
    else:
        img = numpy.zeros( (int(round(511.0 * height / width)), 511), dtype=numpy.uint32 )
    img.shape = (img.shape[1],img.shape[0])
    min_color = (int(min_color[:2], 16) / 255.0, int(min_color[2:4], 16) / 255.0, int(min_color[4:6], 16) / 255.0)
    max_color = (int(max_color[:2], 16) / 255.0, int(max_color[2:4], 16) / 255.0, int(max_color[4:6], 16) / 255.0)
    if mid_color is None:
        mid_color = ((min_color[0] + max_color[0]) / 2.0, (min_color[0] + max_color[0]) / 2.0,
                     (min_color[0] + max_color[0]) / 2.0)
    else:
        mid_color = (int(mid_color[:2], 16) / 255.0, int(mid_color[2:4], 16) / 255.0, int(mid_color[4:6], 16) / 255.0)
    temp0 = numpy.linspace(1.0, 0.0, 256)[:255].astype(numpy.float64).reshape(-1, 1)
    temp1 = 1.0 - temp0
    img[:255, :] = (255 * 256 ** 3.0 + numpy.round(255 * (temp0 * max_color[0] + temp1 * mid_color[0])) +
                    numpy.round(255 * (temp0 * max_color[1] + temp1 * mid_color[1])) * 256 +
                    numpy.round(255 * (temp0 * max_color[2] + temp1 * mid_color[2])) * 256 ** 2).astype(numpy.uint32)
    temp0 = numpy.linspace(0.0, 1.0, 256).astype(numpy.float64).reshape(-1, 1)
    temp1 = 1.0 - temp0
    img[255:, :] = (255 * 256 ** 3.0 + numpy.round(255 * (temp0 * min_color[0] + temp1 * mid_color[0])) +
                    numpy.round(255 * (temp0 * min_color[1] + temp1 * mid_color[1])) * 256 +
                    numpy.round(255 * (temp0 * min_color[2] + temp1 * mid_color[2])) * 256 ** 2).astype(numpy.uint32)
    pilImage = Image.frombuffer( 'RGBA',(img.shape[1],img.shape[0]),img,'raw','RGBA',0,1)
    if orientation in ['left', 'right']:
        c.insert(bitmap.bitmap( 0, 0, pilImage, height=height))
    else:
        c.insert(bitmap.bitmap( 0, 0, pilImage, height=width), [trafo.rotate(-90), trafo.translate(0, height)])
    c.stroke( path.rect( 0, 0, width, height ), [style.linewidth.THin] )
    labels = []
    if num_ticks > 1:
        ticks = numpy.zeros((num_ticks, 4), dtype=numpy.float32)
        if log_display:
            for i in range(num_ticks):
                labels.append(labelformat % exp(min_score + i / (num_ticks - 1.0) * (max_score - min_score)))
        else:
            for i in range(num_ticks):
                labels.append(labelformat % (min_score + i / (num_ticks - 1.0) * (max_score - min_score)))
        if orientation == 'left':
            ypos = numpy.linspace(0.0, height, num_ticks)
            xpos = numpy.ones(num_ticks, dtype=numpy.float32) * -width * 0.5
            ticks[:, 2] = -width * 0.4
            ticks[:, 1] = numpy.linspace(0, height, num_ticks)
            ticks[:, 3] = numpy.linspace(0, height, num_ticks)
            if labelattr is None:
                labelattr = [text.halign.right, text.valign.middle, text.size(-3)]
        elif orientation == 'right':
            ypos = numpy.linspace(0.0, height, num_ticks)
            xpos = numpy.ones(num_ticks, dtype=numpy.float32) * width * 1.5
            ticks[:, 0] = width
            ticks[:, 2] = width * 1.4
            ticks[:, 1] = numpy.linspace(0, height, num_ticks)
            ticks[:, 3] = numpy.linspace(0, height, num_ticks)
            if labelattr is None:
                labelattr = [text.halign.left, text.valign.middle, text.size(-3)]
        elif orientation == 'bottom':
            xpos = numpy.linspace(0.0, width, num_ticks)
            ypos = numpy.ones(num_ticks, dtype=numpy.float32) * -height * 0.5
            ticks[:, 3] = -height * 0.4
            ticks[:, 0] = numpy.linspace(0, width, num_ticks)
            ticks[:, 2] = numpy.linspace(0, width, num_ticks)
            if labelattr is None:
                labelattr = [text.halign.center, text.valign.top, text.size(-3)]
        else:
            xpos = numpy.linspace(0.0, width, num_ticks)
            ypos = numpy.ones(num_ticks, dtype=numpy.float32) * height * 1.5
            ticks[:, 1] = height
            ticks[:, 3] = height * 1.4
            ticks[:, 0] = numpy.linspace(0, width, num_ticks)
            ticks[:, 2] = numpy.linspace(0, width, num_ticks)
            if labelattr is None:
                labelattr = [text.halign.center, text.valign.bottom, text.size(-3)]
        for i in range(num_ticks):
            c.stroke( path.line(ticks[i, 0], ticks[i, 1], ticks[i, 2], ticks[i, 3]), [style.linewidth.THin] )
            c.text(xpos[i], ypos[i], r"%s" % labels[i], labelattr)
    return c



# distutils: language = c++

"""These functions provide increased speed in HMM processes.
"""

import cython
cimport numpy as np
import numpy

ctypedef np.float32_t DTYPE_t
ctypedef np.float64_t DTYPE_64_t
ctypedef np.int32_t DTYPE_int_t
ctypedef np.int64_t DTYPE_int64_t
ctypedef np.uint32_t DTYPE_uint_t
ctypedef np.int8_t DTYPE_int8_t
cdef double Inf = numpy.inf
cdef double PI = numpy.pi

cdef extern from "math.h":
    double exp(double x) nogil
    double log(double x) nogil
    double log2(double x) nogil
    double log10(double x) nogil
    double sqrt(double x) nogil
    double pow(double x, double x) nogil
    double abs(double x) nogil
    double round(double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_probabilities(
        np.ndarray[DTYPE_64_t, ndim=1] observations,
        np.ndarray[DTYPE_64_t, ndim=3] distributions,
        np.ndarray[DTYPE_64_t, ndim=3] probabilities):
    cdef int i, j, k
    cdef double temp, prob, sigma2, scalar
    cdef int num_states = probabilities.shape[0]
    cdef int num_dists = probabilities.shape[1] - 1
    cdef int num_obs = probabilities.shape[2]
    with nogil:
        for i in range(num_states):
            for j in range(num_dists):
                scalar = distributions[i, j, 0] * pow(2.0 * PI * distributions[i, j, 2], -0.5)
                sigma2 = 1.0 / (2.0 * distributions[i, j, 2])
                for k in range(num_obs):
                    temp = observations[k] - distributions[i, j, 1]
                    prob = scalar * exp(-temp * temp * sigma2)
                    probabilities[i, j, k] = prob
                    probabilities[i, num_dists, k] += prob
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_alphas(
        np.ndarray[DTYPE_64_t, ndim=3] probabilities,
        np.ndarray[DTYPE_64_t, ndim=1] pi,
        np.ndarray[DTYPE_64_t, ndim=2] transitions,
        np.ndarray[DTYPE_64_t, ndim=2] alphas,
        np.ndarray[DTYPE_64_t, ndim=1] scalars):
    cdef int i, j, k
    cdef int num_states = probabilities.shape[0]
    cdef int num_dists = probabilities.shape[1] - 1
    cdef int num_obs = probabilities.shape[2]
    with nogil:
        for i in range(num_states):
            alphas[i, 0] = probabilities[i, num_dists, 0] * pi[i]
            scalars[0] += alphas[i, 0]
        for i in range(num_states):
            alphas[i, 0] /= scalars[0]
        for k in range(1, num_obs):
            for i in range(num_states):
                for j in range(num_states):
                    alphas[i, k] += alphas[j, k - 1] * transitions[j, i]
                alphas[i, k] *= probabilities[i, num_dists, k]
                scalars[k] += alphas[i, k]
            for i in range(num_states):
                alphas[i, k] /= scalars[k]
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_betas(
        np.ndarray[DTYPE_64_t, ndim=3] probabilities,
        np.ndarray[DTYPE_64_t, ndim=2] transitions,
        np.ndarray[DTYPE_64_t, ndim=1] scalars,
        np.ndarray[DTYPE_64_t, ndim=2] betas):
    cdef int i, j, k, l
    cdef double total, temp
    cdef int num_states = probabilities.shape[0]
    cdef int num_dists = probabilities.shape[1] - 1
    cdef int num_obs = probabilities.shape[2]
    with nogil:
        total = 0.0
        for i in range(num_states):
            betas[i, num_obs - 1] = 1.0
        for l in range(1, num_obs):
            k = num_obs - l - 1
            for i in range(num_states):
                for j in range(num_states):
                    betas[i, k] += betas[j, k + 1] * probabilities[j, num_dists, k + 1] * transitions[i, j]
                betas[i, k] /= scalars[k + 1]
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_etas(
        np.ndarray[DTYPE_64_t, ndim=3] probabilities,
        np.ndarray[DTYPE_64_t, ndim=2] transitions,
        np.ndarray[DTYPE_64_t, ndim=2] alphas,
        np.ndarray[DTYPE_64_t, ndim=2] betas,
        np.ndarray[DTYPE_64_t, ndim=3] etas):
    cdef int i, j, k, l
    cdef double total, temp
    cdef int num_states = probabilities.shape[0]
    cdef int num_dists = probabilities.shape[1] - 1
    cdef int num_obs = probabilities.shape[2]
    with nogil:
        for k in range(num_obs - 1):
            total = 0.0
            for i in range(num_states):
                for j in range(num_states):
                    etas[i, j, k] = (alphas[i, k] * transitions[i, j] * probabilities[j, num_dists, k + 1] *
                                     betas[j, k + 1])
                    total += etas[i, j, k]
            for i in range(num_states):
                for j in range(num_states):
                    etas[i, j, k] /= total
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_gammas(
        np.ndarray[DTYPE_64_t, ndim=3] probabilities,
        np.ndarray[DTYPE_64_t, ndim=2] alphas,
        np.ndarray[DTYPE_64_t, ndim=2] betas,
        np.ndarray[DTYPE_64_t, ndim=3] gammas):
    cdef int i, j, k, l
    cdef double total, temp
    cdef int num_states = probabilities.shape[0]
    cdef int num_dists = probabilities.shape[1] - 1
    cdef int num_obs = probabilities.shape[2]
    with nogil:
        for k in range(num_obs):
            for i in range(num_states):
                temp = alphas[i, k] * betas[i, k] / probabilities[i, num_dists, k]
                for j in range(num_dists):
                    gammas[i, j, k] = temp * probabilities[i, j, k]
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def update_parameter_estimates(
        np.ndarray[DTYPE_64_t, ndim=1] observations,
        np.ndarray[DTYPE_64_t, ndim=1] pi,
        np.ndarray[DTYPE_64_t, ndim=2] transitions,
        np.ndarray[DTYPE_64_t, ndim=3] distributions,
        np.ndarray[DTYPE_64_t, ndim=1] new_pi,
        np.ndarray[DTYPE_64_t, ndim=2] new_transitions,
        np.ndarray[DTYPE_64_t, ndim=3] new_distributions,
        np.ndarray[DTYPE_64_t, ndim=3] etas,
        np.ndarray[DTYPE_64_t, ndim=3] gammas,
        int finalize):
    cdef int i, j, k
    cdef double total, temp, change
    cdef int num_states = gammas.shape[0]
    cdef int num_dists = gammas.shape[1]
    cdef int num_obs = gammas.shape[2]
    with nogil:
        change = 0.0
        for i in range(num_states):
            for j in range(num_states):
                new_pi[i] += etas[i, j, 0]
        for k in range(num_obs - 1):
            for i in range(num_states):
                for j in range(num_states):
                    new_transitions[i, j] += etas[i, j, k]
        for k in range(num_obs):
            for i in range(num_states):
                for j in range(num_dists):
                    new_distributions[i, j, 0] += gammas[i, j, k] 
                    new_distributions[i, j, 1] += gammas[i, j, k] * observations[k]
                    temp = observations[k] - distributions[i, j, 1]
                    new_distributions[i, j, 2] += gammas[i, j, k] * temp * temp
        if finalize == 1:
            total = 0.0
            for i in range(num_states):
                total += new_pi[i]
            for i in range(num_states):
                new_pi[i] /= total
                temp = new_pi[i] - pi[i]
                if temp > 0.0:
                    change = max(change, temp)
                else:
                    change = max(change, -temp)
                pi[i] = new_pi[i]
            for i in range(num_states):
                total = 0.0
                for j in range(num_states):
                    total += new_transitions[i, j]
                for j in range(num_states):
                    new_transitions[i, j] /= total
                    temp = new_transitions[i, j] - transitions[i, j]
                    if temp > 0.0:
                        change = max(change, temp)
                    else:
                        change = max(change, -temp)
                    transitions[i, j] = new_transitions[i, j]
            for i in range(num_states):
                total = 0.0
                for j in range(num_dists):
                    total += new_distributions[i, j, 0]
                    if new_distributions[i, j, 0] > 0:
                        new_distributions[i, j, 1] /= new_distributions[i, j, 0]
                        new_distributions[i, j, 2] /= new_distributions[i, j, 0]
                for j in range(num_dists):
                    if new_distributions[i, j, 0] > 0:
                        new_distributions[i, j, 0] /= total
                        temp = new_distributions[i, j, 0] - distributions[i, j, 0]
                        if temp > 0.0:
                            change = max(change, temp)
                        else:
                            change = max(change, -temp)
                        distributions[i, j, 0] = new_distributions[i, j, 0]
                        temp = new_distributions[i, j, 1] - distributions[i, j, 1]
                        if temp > 0.0:
                            change = max(change, temp)
                        else:
                            change = max(change, -temp)
                        distributions[i, j, 1] = new_distributions[i, j, 1]
                        temp = new_distributions[i, j, 2] - distributions[i, j, 2]
                        if temp > 0.0:
                            change = max(change, temp)
                        else:
                            change = max(change, -temp)
                        distributions[i, j, 2] = new_distributions[i, j, 2]
    return change

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_path(
        np.ndarray[DTYPE_64_t, ndim=1] observations,
        np.ndarray[DTYPE_64_t, ndim=1] pi,
        np.ndarray[DTYPE_64_t, ndim=2] transitions,
        np.ndarray[DTYPE_64_t, ndim=2] probs,
        np.ndarray[DTYPE_64_t, ndim=2] scores,
        np.ndarray[DTYPE_int_t, ndim=2] paths,
        np.ndarray[DTYPE_int_t, ndim=1] states):
    cdef int i, j, k
    cdef double prob
    cdef int num_states = probs.shape[0]
    cdef int num_obs = probs.shape[1]
    with nogil:
        for i in range(num_states):
            scores[i, 0] = probs[i, 0] + pi[i]
        for k in range(1, num_obs):
            for i in range(num_states):
                scores[i, k] = -Inf
                for j in range(num_states):
                    prob = scores[j, k - 1] + transitions[j, i]
                    if prob > scores[i, k]:
                        scores[i, k] = prob
                        paths[i, k - 1] = j
                scores[i, k] += probs[i, k]
        prob = scores[0, num_obs - 1]
        states[num_obs - 1] = 0
        for i in range(1, num_states):
            if scores[i, num_obs - 1] > prob:
                prob = scores[i, num_obs - 1]
                states[num_obs - 1] = i
        for i in range(2, num_obs + 1):
            j = num_obs - i
            states[j] = paths[states[j + 1], j]
    return prob

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def generate_sequence(
        np.ndarray[DTYPE_64_t, ndim=2] rand_nums,
        np.ndarray[DTYPE_64_t, ndim=1] pi,
        np.ndarray[DTYPE_64_t, ndim=2] transitions,
        np.ndarray[DTYPE_64_t, ndim=3] distributions,
        np.ndarray[DTYPE_64_t, ndim=1] pi_sums,
        np.ndarray[DTYPE_64_t, ndim=2] transition_sums,
        np.ndarray[DTYPE_64_t, ndim=2] weight_sums,
        np.ndarray[DTYPE_int_t, ndim=1] states,
        np.ndarray[DTYPE_int_t, ndim=1] distribution):
    cdef int i, j, state, dist
    cdef int num_obs = states.shape[0]
    cdef int num_states = pi.shape[0]
    cdef int num_dists = distributions.shape[1]
    with nogil:
        pi_sums[0] = pi[0]
        for i in range(1, num_states):
            pi_sums[i] = pi_sums[i - 1] + pi[i]
        for i in range(num_states):
            transition_sums[i, 0] = transitions[i, 0]
            for j in range(1, num_states):
                transition_sums[i, j] = transition_sums[i, j - 1] + transitions[i, j]
            weight_sums[i, 0] = distributions[i, 0, 0]
            for j in range(1, num_dists):
                weight_sums[i, j] = weight_sums[i, j - 1] + distributions[i, j, 0]
        state = 0
        while rand_nums[0, 0] > pi_sums[state]:
            state += 1
        states[0] = state
        dist = 0
        while rand_nums[0, 1] > weight_sums[state, dist]:
            dist += 1
        distribution[0] = dist
        for i in range(1, num_obs):
            state = 0
            while rand_nums[i, 0] > transition_sums[states[i - 1], state]:
                state += 1
            states[i] = state
            dist = 0
            while rand_nums[i, 1] > weight_sums[state, dist]:
                dist += 1
            distribution[i] = dist
    return None







# distutils: language = c++

"""These functions provide increased speed in handling functions dealing with finding valid
interactions and interaction ranges necessary for supporting hic analysis using HiFive.
"""

import cython
cimport numpy as np
import numpy

ctypedef np.float32_t DTYPE_t
ctypedef np.float64_t DTYPE_64_t
ctypedef np.int32_t DTYPE_int_t
ctypedef np.int64_t DTYPE_int64_t
ctypedef np.uint32_t DTYPE_uint_t
ctypedef np.int8_t DTYPE_int8_t
cdef double Inf = numpy.inf

cdef extern from "math.h":
    double exp(double x) nogil
    double log(double x) nogil
    double log10(double x) nogil
    double sqrt(double x) nogil
    double pow(double x, double x) nogil
    double abs(double x) nogil
    double round(double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_max_bin(
        np.ndarray[DTYPE_int_t, ndim=2] binbounds not None,
        int maxdistance):
    cdef int i, j, maxbin, start
    cdef int num_bins = binbounds.shape[0]
    with nogil:
        maxbin = 0
        for i in range(num_bins):
            start = binbounds[i, 1] - 1
            j = i + 1
            while j < num_bins and binbounds[j, 0] - start < maxdistance:
                j += 1
            maxbin = max(maxbin, j - i - 1)
    return maxbin

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_max_fend(
        np.ndarray[DTYPE_int_t, ndim=1] max_fend not None,
        np.ndarray[DTYPE_int_t, ndim=1] mids not None,
        np.ndarray[DTYPE_int_t, ndim=1] chromosomes not None,
        np.ndarray[DTYPE_int_t, ndim=1] chr_indices not None,
        int start,
        int maxdistance):
    cdef long long int h, i, j, chr_max
    cdef long long int num_fends = max_fend.shape[0]
    cdef long long int max_num_fends = mids.shape[0]
    with nogil:
        j = 1
        for i in range(num_fends):
            chr_max = min(max_num_fends, chr_indices[chromosomes[i] + 1] - start)
            if maxdistance == 0:
                max_fend[i] = chr_max
            else:
                while j < chr_max and mids[j] - mids[i] < maxdistance:
                    j += 1
                max_fend[i] = j
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_min_fend(
        np.ndarray[DTYPE_int_t, ndim=1] min_fend not None,
        np.ndarray[DTYPE_int_t, ndim=1] mids not None,
        np.ndarray[DTYPE_int_t, ndim=1] chromosomes not None,
        np.ndarray[DTYPE_int_t, ndim=1] chr_indices not None,
        int mindistance,
        int binned):
    cdef long long int h, i, j, chr_max
    cdef long long int num_fends = min_fend.shape[0]
    cdef long long int max_num_fends = mids.shape[0]
    with nogil:
        if binned == 1:
            j = 0
        else:
            j = 2
        for i in range(num_fends):
            chr_max = min(max_num_fends, chr_indices[chromosomes[i] + 1])
            if mindistance == 0:
                if binned == 1:
                    min_fend[i] = min(i, chr_max)
                else:
                    min_fend[i] = min(i + 2, chr_max)
            else:
                while j < chr_max and mids[j] - mids[i] < mindistance:
                    j += 1
                min_fend[i] = j
    return None

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_fend_coverage(
        np.ndarray[DTYPE_int_t, ndim=2] data not None,
        np.ndarray[DTYPE_int64_t, ndim=1] data_indices not None,
        np.ndarray[DTYPE_int_t, ndim=1] filter not None,
        np.ndarray[DTYPE_int_t, ndim=1] min_fend not None,
        np.ndarray[DTYPE_int_t, ndim=1] max_fend not None,
        np.ndarray[DTYPE_int_t, ndim=1] coverage not None,
        int mincoverage):
    cdef long long int i, fend1, valid
    cdef long long int num_fends = filter.shape[0]
    with nogil:
        valid = 0
        for fend1 in range(num_fends):
            if filter[fend1] == 0:
                continue
            i = data_indices[fend1]
            while i < data_indices[fend1 + 1] and data[i, 1] < min_fend[fend1]:
                i += 1
            while i < data_indices[fend1 + 1] and data[i, 1] < max_fend[fend1]:
                if filter[data[i, 1]] == 1:
                    coverage[fend1] += 1
                    if fend1 != data[i, 1]:
                        coverage[data[i, 1]] += 1
                i += 1
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_fend_ranges(
        np.ndarray[DTYPE_int_t, ndim=1] mapping not None,
        np.ndarray[DTYPE_int_t, ndim=1] mids not None,
        np.ndarray[DTYPE_int64_t, ndim=2] ranges not None,
        int mindistance,
        int maxdistance,
        long long int start,
        long long int stop,
        int startfend,
        int binned):
    cdef long long int i, j, pos, total
    cdef long long int total_fends = mapping.shape[0]
    with nogil:
        for i in range(start, stop):
            pos = mids[i]
            j = i + 1 - binned
            while j < total_fends and mids[j] - pos < mindistance:
                j += 1
            ranges[i, 1] = j
            j = total_fends
            while j > ranges[i, 1] and mids[j - 1] - pos > maxdistance:
                j -= 1
            ranges[i, 2] = j
            total = ranges[i, 2] - ranges[i, 1]
            if binned == 0:
                if ranges[i, 1] < total_fends and mapping[ranges[i, 1]] - mapping[i] == 1:
                    total -= 1
                if (mapping[i] + startfend) % 2 == 0:
                    if ranges[i, 1] < total_fends and mapping[ranges[i, 1]] - mapping[i] == 3:
                        total -= 1
                    if ranges[i, 1] + 1 < total_fends and mapping[ranges[i, 1] + 1] - mapping[i] == 3:
                        total -= 1
                    if ranges[i, 1] + 2 < total_fends and mapping[ranges[i, 1] + 2] - mapping[i] == 3:
                        total -= 1
            ranges[i, 0] = total
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_nonzeros_in_range(
        np.ndarray[DTYPE_int64_t, ndim=2] ranges not None,
        np.ndarray[DTYPE_int_t, ndim=2] data not None):
    cdef long long int i, count
    cdef long long int num_data = data.shape[0]
    with nogil:
        count = 0
        for i in range(num_data):
            if data[i, 0] == -1:
                continue
            if data[i, 1] >= ranges[data[i, 0], 1] and data[i, 1] < ranges[data[i, 0], 2]:
                count += 1
    return count


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_nonzero_node_indices(
        np.ndarray[DTYPE_int64_t, ndim=2] ranges,
        np.ndarray[DTYPE_int_t, ndim=1] indices0,
        np.ndarray[DTYPE_int_t, ndim=1] indices1,
        np.ndarray[DTYPE_int_t, ndim=1] counts,
        np.ndarray[DTYPE_int_t, ndim=2] data,
        np.ndarray[DTYPE_int_t, ndim=1] interactions):
    cdef long long int i, j
    cdef long long int num_data = data.shape[0]
    with nogil:
        i = 0
        for j in range(num_data):
            if data[j, 0] == -1:
                continue
            if data[j, 1] >= ranges[data[j, 0], 1] and data[j, 1] < ranges[data[j, 0], 2]:
                indices0[i] = data[j, 0]
                indices1[i] = data[j, 1]
                interactions[data[j, 0]] += 1
                if data[j, 0] != data[j, 1]:
                    interactions[data[j, 1]] += 1
                if not counts is None:
                    counts[i] = data[j, 2]
                i += 1
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_zero_node_indices(
        np.ndarray[DTYPE_int_t, ndim=1] mapping,
        np.ndarray[DTYPE_int64_t, ndim=2] ranges,
        np.ndarray[DTYPE_int_t, ndim=1] nzindices0,
        np.ndarray[DTYPE_int_t, ndim=1] nzindices1,
        np.ndarray[DTYPE_int_t, ndim=1] zindices0,
        np.ndarray[DTYPE_int_t, ndim=1] zindices1,
        np.ndarray[DTYPE_int_t, ndim=1] interactions,
        long long int start,
        long long int stop,
        int startfend,
        int binned):
    cdef long long int i, j, m1, m2, nzpos, zpos
    cdef long long int num_nz = nzindices0.shape[0]
    with nogil:
        nzpos = 0
        zpos = 0
        for i in range(start, stop):
            while nzpos < num_nz and nzindices0[nzpos] < i:
                nzpos += 1
            for j in range(ranges[i, 1], ranges[i, 2]):
                while nzpos < num_nz and nzindices0[nzpos] == i and nzindices1[nzpos] < j:
                    nzpos += 1
                m1 = mapping[i]
                m2 = mapping[j]
                if binned == 0 and m2 - m1 < 4 and (m2 - m1 == 1 or ((m1 + startfend) % 2 == 0 and m2 - m1 == 3)):
                    continue
                elif nzpos == num_nz or nzindices1[nzpos] != j or nzindices0[nzpos] != i:
                    zindices0[zpos] = i
                    zindices1[zpos] = j
                    interactions[i] += 1
                    if i != j:
                        interactions[j] += 1
                    zpos += 1
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_distancebound_possible_interactions(
        np.ndarray[DTYPE_int64_t, ndim=1] interactions not None,
        np.ndarray[DTYPE_int_t, ndim=1] chr_indices not None,
        np.ndarray[DTYPE_int_t, ndim=1] mids not None,
        np.ndarray[DTYPE_int_t, ndim=1] filter not None,
        np.ndarray[DTYPE_int_t, ndim=1] chrints not None,
        int useread_int,
        int mindistance,
        int maxdistance,
        int binned):
    cdef long long int h, i, j, k, chr_total, all_total, total, stop
    cdef long long int num_chroms = chrints.shape[0]
    cdef long long int num_fends = filter.shape[0]
    with nogil:
        all_total = 0
        chr_total = 0
        # if trans interactions are included, find total valid fends
        if useread_int != 0:
            for i in range(num_fends):
                all_total += filter[i]
        for h in range(num_chroms):
            i = chrints[h]
            # for each chrom, find valid number of fends
            chr_total = 0
            for j in range(chr_indices[i], chr_indices[i + 1]):
                chr_total += filter[j]
            # if using trans, for each valid fend add on all possible trans interactions tot total
            if useread_int > 0:
                for j in range(chr_indices[i], chr_indices[i + 1]):
                    if filter[j] > 0:
                        interactions[j] += all_total - chr_total
            # if using cis, find number of interactions within distance range and
            # not adjacent fends or opposite strand adjacent fragment fends
            if useread_int < 2:
                for j in range(chr_indices[i], chr_indices[i + 1]):
                    if filter[j] == 0:
                        continue
                    if binned == 0:
                        total = chr_total - 1
                        if j > chr_indices[i] + 4:
                            # remove upstream interactions outside of maxdistance
                            k = chr_indices[i]
                            stop = (j / 2) * 2 - 2
                            while k < stop and mids[j] - mids[k] >= maxdistance:
                                total -= filter[k]
                                k += 1
                            stop = k - 1
                            # remove upstream interactions inside of mindistance
                            k = (j / 2) * 2 - 3
                            while k > stop and mids[j] - mids[k] < mindistance:
                                total -= filter[k]
                                k -= 1
                        if j < chr_indices[i + 1] - 4:
                            # remove downstream interactions outside of maxdistance
                            k = chr_indices[i + 1] - 1
                            stop = (j / 2) * 2 + 3
                            while k > stop and mids[k] - mids[j] >= maxdistance:
                                total -= filter[k]
                                k -= 1
                            stop = k + 1
                            # remove downstream interactions inside of mindistance
                            k = (j / 2) * 2 + 4
                            while k < stop and mids[k] - mids[j] < mindistance:
                                total -= filter[k]
                                k += 1
                        if j % 2 == 0:
                            # remove same fragment fend
                            total -= filter[j + 1]
                            # remove previous fragment fends if necessary
                            if j > chr_indices[i]:
                                total -= filter[j - 1]
                                if mids[j] - mids[j - 2] < mindistance:
                                    total -= filter[j - 2]
                            # remove next fragment fends if necessary
                            if j < chr_indices[i + 1] - 2:
                                total -= filter[j + 3]
                                if mids[j + 2] - mids[j] < mindistance:
                                    total -= filter[j + 2]
                        else:
                            # remove same fragment fend
                            total -= filter[j - 1]
                            # remove previous fragment fends if necessary
                            if j > chr_indices[i] + 2:
                                total -= filter[j - 3]
                                if mids[j] - mids[j - 2] < mindistance:
                                    total -= filter[j - 2]
                            # remove next fragment fends if necessary
                            if j < chr_indices[i + 1] - 2:
                                total -= filter[j + 1]
                                if mids[j + 2] - mids[j] < mindistance:
                                    total -= filter[j + 2]
                    else:
                        total = chr_total
                        if mindistance > 0:
                            total -= 1
                        # remove upstream interactions outside of maxdistance
                        k = chr_indices[i]
                        while k < j and mids[j] - mids[k] >= maxdistance:
                            total -= filter[k]
                            k += 1
                        stop = k - 1
                        # remove upstream interactions inside of mindistance
                        k = j - 1
                        while k > stop and mids[j] - mids[k] < mindistance:
                            total -= filter[k]
                            k -= 1
                        # remove downstream interactions outside of maxdistance
                        k = chr_indices[i + 1] - 1
                        while k > j and mids[k] - mids[j] >= maxdistance:
                            total -= filter[k]
                            k -= 1
                        stop = k + 1
                        # remove downstream interactions inside of mindistance
                        k = j + 1
                        while k < stop and mids[k] - mids[j] < mindistance:
                            total -= filter[k]
                            k += 1
                    interactions[j] += total
    return None


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def sum_weighted_indices(
        np.ndarray[DTYPE_int_t, ndim=1] indices0,
        np.ndarray[DTYPE_int_t, ndim=1] indices1,
        np.ndarray[DTYPE_t, ndim=1] weights,
        np.ndarray[DTYPE_int_t, ndim=1] int_weights,
        np.ndarray[DTYPE_64_t, ndim=1] sums):
    cdef long long int i
    cdef long long int num_indices = indices0.shape[0]
    with nogil:
        if not weights is None:
            for i in range(num_indices):
                sums[indices0[i]] += weights[i]
                sums[indices1[i]] += weights[i]
        elif not int_weights is None:
            for i in range(num_indices):
                sums[indices0[i]] += int_weights[i]
                sums[indices1[i]] += int_weights[i]
        else:
            for i in range(num_indices):
                sums[indices0[i]] += 1
                sums[indices1[i]] += 1
    return None







#!/usr/bin/env python

"""A class for handling HiC modeling."""

import os
import sys

import numpy
import mlpy

from hic_binning import dynamically_bin_cis_array
import libraries._modeling as _modeling

class ThreeDModel(object):
    """
    This class handles the creation of 3D models based on HiC data, including learning parameters,
    exporting coordinates into JSON format.
    """

    def __init__(self, hic, chrom, start=None, stop=None, binsize=10000, method='poisson', max_iterations=1000,
                 silent=False):
        """Create a 3D model using the specified algorithmic approach."""
        self.silent = silent
        if chrom not in hic.chr2int:
            if not self.silent:
                print >> sys.stderr, ("Chromosome name not found in HiC project.\n"),
            return None
        self.method = method
        if method not in ['poisson']:
            if not self.silent:
                print >> sys.stderr, ("Method name not recognized.\n"),
            return None
        chrint = hic.chr2int[chrom]
        start_fend = hic.fends['chr_indices'][chrint]
        stop_fend = hic.fends['chr_indices'][chrint]
        if start is None:
            while start_fend < stop_fend and hic.filter[start_fend] == 0:
                start_fend += 1
            start = hic.fends['fends']['mid'][start_fend]
        if stop is None:
            while stop_fend > start_fend and hic.filter[stop_fend - 1] == 0:
                stop_fend -= 1
        self.start = start
        self.stop = stop
        self.chrom = chrom
        self.binsize = binsize
        if method == 'poisson':
            self._learn_poisson_model(hic, max_iterations)
        return None

    def _learn_poisson_model(self, hic, max_iterations=1000):
        data, mapping = hic.cis_heatmap(self.chrom, start=self.start, stop=self.stop, binsize=self.binsize,
                                               datatype='fend', arraytype='upper', returnmapping=True)
        self.positions = (mapping[:, 0] + mapping[:, 1]) / 2
        binned_data = numpy.copy(data)
        dynamically_bin_cis_array(data, mapping, binned_data, mapping, minobservations=10)
        distances = numpy.zeros((mapping.shape[0], mapping.shape[0]), dtype=numpy.float32)
        indices = numpy.triu_indices(mapping.shape[0], 1)
        """
        distances[indices] = (binned_data[:, 0] / binned_data[:, 1]) ** -2.0
        distances[indices[1], indices[0]] = distances[indices]
        if not self.silent:
            print >> sys.stderr, ('Finding initial coordinate states...'),
        #coords = self.PCA(distances, 3)
        U, s, Vt = svd(distances, full_matrices=False)
        order = numpy.argsort(s)[::-1]
        U = U[:, order]
        s = s[order]
        Vt = Vt[:, order]
        PCAs = ((s**0.5) * Vt.T)[:3, :]
        coords = PCAs.dot(distances).T
        print coords.shape
        """
        distances[indices] = numpy.log(binned_data[:, 0] / binned_data[:, 1])
        distances[indices[1], indices[0]] = distances[indices]
        distances[numpy.arange(distances.shape[0]), numpy.arange(distances.shape[0])] = numpy.amax(distances[indices])
        distances -= numpy.mean(distances, axis=1).reshape(-1, 1)
        distances /= numpy.sqrt(numpy.sum(distances**2, axis=1)).reshape(-1, 1)
        pca_fast = mlpy.PCAFast(k=3)
        pca_fast.learn(distances)
        coords = pca_fast.transform(distances).astype(numpy.float32)
        if not self.silent:
            print >> sys.stderr, ('Done\nLearning coordinates...'),
        del binned_data
        distances = distances[indices]
        nonzero_indices = numpy.where(data[:, 0] > 0)[0].astype(numpy.int32)
        nonzero_indices0 = indices[0][nonzero_indices].astype(numpy.int32)
        nonzero_indices1 = indices[1][nonzero_indices].astype(numpy.int32)
        nonzero_counts = data[nonzero_indices, 0].astype(numpy.int32)
        nonzero_weights = data[nonzero_indices, 1].astype(numpy.float32)
        zero_indices = numpy.where(data[:, 0] == 0)[0].astype(numpy.int32)
        zero_indices0 = indices[0][zero_indices].astype(numpy.int32)
        zero_indices1 = indices[1][zero_indices].astype(numpy.int32)
        zero_weights = data[zero_indices, 1].astype(numpy.float32)
        del indices
        alpha = -3.0
        beta = 0.1
        new_coords = numpy.copy(coords)
        coord_grads = numpy.zeros(coords.shape, dtype=numpy.float64)
        param_grads = numpy.zeros(2, dtype=numpy.float64)
        _modeling.find_distances(coords, distances)
        nonzero_log_distances = numpy.log(distances[nonzero_indices]).astype(numpy.float32)
        start_cost = _modeling.find_poisson_cost(zero_indices,
                                                 zero_weights,
                                                 nonzero_indices,
                                                 nonzero_weights,
                                                 nonzero_counts,
                                                 nonzero_log_distances,
                                                 distances,
                                                 alpha,
                                                 beta)
        print start_cost
        cost = start_cost
        previous_cost = start_cost
        change = 0.0
        if max_iterations > 0:
            cont = True
        else:
            cont = False
        iteration = 0
        learningstep = 0.5
        while cont:
            """
            param_grads.fill(0.0)
            zero_log_distances = numpy.log(distances[zero_indices]).astype(numpy.float32)
            _modeling.calculate_poisson_param_gradients(zero_indices,
                                                        zero_weights,
                                                        zero_log_distances,
                                                        nonzero_indices,
                                                        nonzero_weights,
                                                        nonzero_counts,
                                                        nonzero_log_distances,
                                                        distances,
                                                        param_grads,
                                                        alpha,
                                                        beta)
            gradient_norm = numpy.sum(param_grads ** 2.0)
            # find best step size
            armijo = numpy.inf
            t = 1.0
            while armijo > 0.0:
                # if using multiple cores, pass gradients to root
                new_alpha = alpha - t * param_grads[0] / distances.shape[0]
                new_beta = beta - t * param_grads[1] / distances.shape[0]
                print param_grads, new_alpha, new_beta
                cost = _modeling.find_poisson_cost(zero_indices,
                                                   zero_weights,
                                                   nonzero_indices,
                                                   nonzero_weights,
                                                   nonzero_counts,
                                                   nonzero_log_distances,
                                                   distances,
                                                   new_alpha,
                                                   new_beta)
                armijo = cost - previous_cost + t * gradient_norm
                if not self.silent:
                    print >> sys.stderr, ("\r%s iteration:%ia cost:%f change:%f armijo: %f %s") %\
                                         ('Learning coordinates...', iteration, previous_cost,
                                          change, armijo, ' ' * 20),
                t *= learningstep
            previous_cost = cost
            alpha = new_alpha
            beta = new_beta
            if not self.silent:
                print >> sys.stderr, ("\r%s iteration:%ia cost:%f %s") %\
                                     ('Learning coordinates...', iteration, cost, ' ' * 40),
            """
            coord_grads.fill(0.0)
            _modeling.calculate_poisson_coord_gradients(zero_indices,
                                                        zero_indices0,
                                                        zero_indices1,
                                                        zero_weights,
                                                        nonzero_indices,
                                                        nonzero_indices0,
                                                        nonzero_indices1,
                                                        nonzero_weights,
                                                        nonzero_counts,
                                                        distances,
                                                        coords,
                                                        coord_grads,
                                                        alpha,
                                                        beta)
            gradient_norm = numpy.sum(coord_grads ** 2.0)
            print numpy.amin(numpy.abs(coord_grads)), numpy.amax(numpy.abs(coord_grads))
            # find best step size
            armijo = numpy.inf
            t = 1.0
            while armijo > 0.0:
                # if using multiple cores, pass gradients to root
                new_coords = (coords - t * coord_grads).astype(numpy.float32)
                print numpy.sum(numpy.abs(new_coords - coords))
                _modeling.find_distances(new_coords, distances)
                nonzero_log_distances = numpy.log(distances[nonzero_indices]).astype(numpy.float32)
                cost = _modeling.find_poisson_cost(zero_indices,
                                                   zero_weights,
                                                   nonzero_indices,
                                                   nonzero_weights,
                                                   nonzero_counts,
                                                   nonzero_log_distances,
                                                   distances,
                                                   alpha,
                                                   beta)
                armijo = cost - previous_cost + t * gradient_norm
                if not self.silent:
                    print >> sys.stderr, ("\r%s iteration:%ib cost:%f change:%f armijo: %f %s") %\
                                         ('Learning coodrinates...', iteration, previous_cost,
                                          change, armijo, ' ' * 20),
                t *= learningstep
                armijo = -numpy.inf
            previous_cost = cost
            coords = new_coords
            if not self.silent:
                print >> sys.stderr, ("\r%s iteration:%ib cost:%f %s") %\
                                     ('Learning cooordinates...', iteration, cost, ' ' * 40),
            iteration += 1
            if iteration >= max_iterations:
                cont = False
        self.coords = coords
        self.alpha = alpha
        self.beta = beta
        self.coords /= numpy.mean(distances) / 100.0
        self.coords -= numpy.mean(self.coords, axis=0).reshape(1, -1)
        if not self.silent:
            print >> sys.stderr, ("\r%s\tLearning coordinates... Final cost: %f\n") % (" " * 80, cost),
        return None

    def PCA(self, data, k):
        assert len(data.shape) == 2, "Data must be 2D"
        m, n = data.shape
        means = numpy.mean(data, axis=0)
        scatter_matrix = numpy.sum((data.reshape(m, 1, n) - means.reshape(1, 1, n)) *
                                   (data.reshape(m, n, 1) - means.reshape(1, n, 1)), axis=0)
        eig_val, eig_vec = numpy.linalg.eig(scatter_matrix)
        order = numpy.argsort(eig_val)[::-1]
        eig_vec = eig_vec[:, order[:k]]
        new_data = eig_vec.T.dot(data)
        return new_data.astype(numpy.float32).T

    def export_json(self, fname):
        output = open(fname, 'w')
        print >> output, "var DNA = {"
        print >> output, "\tnum_points:%i," % self.positions.shape[0]
        print >> output, "\tpositions:["
        for i in range(self.coords.shape[0]):
            if i < self.coords.shape[0] - 1:
                print >> output, "\t\t%i," % self.positions[i]
            else:
                print >> output, "\t\t%i" % self.positions[i]
        print >> output, "\t\t],\n\tcoordinates:["
        for i in range(self.coords.shape[0]):
            if i < self.coords.shape[0] - 1:
                print >> output, "\t\tnew THREE.Vector3(%f, %f, %f)," % (self.coords[i, 0],
                                 self.coords[i, 1], self.coords[i, 2])
            else:
                print >> output, "\t\tnew THREE.Vector3(%f, %f, %f)" % (self.coords[i, 0],
                                 self.coords[i, 1], self.coords[i, 2])
        print >> output, "\t]\n};"
        output.close()
        return None



#!/usr/bin/env python

import sys
import os

import h5py


def run(args):
    target_fname = os.path.abspath(args.target)
    source_fname = os.path.abspath(args.source)
    infile = h5py.File(target_fname, 'r')
    if args.type in ['hic-data', '5c-data']:
        if args.type == 'hic-data':
            if 'fendfilename' in infile['/'].attrs:
                secondary_fname = infile['/'].attrs['fendfilename']
            else:
                print >> sys.stderr, ('Unable to find fend filename in data file. The fend file link will not be updated.\n'),
                secondary_fname = None
        elif args.type == '5c-data':
            if 'fragfilename' in infile['/'].attrs:
                secondary_fname = infile['/'].attrs['fragfilename']
            else:
                print >> sys.stderr, ('Unable to find fragment filename in data file. The fragment file link will not be updated.\n'),
                secondary_fname = None
    else:
        secondary_fname = None
    print 'source: %s' % source_fname
    print 'target: %s' % target_fname
    print 'secondary:', secondary_fname
    if secondary_fname is not None:
        if secondary_fname[:2] == './':
            secondary_fname = secondary_fname[2:]
        parent_count = secondary_fname.count('../')
        secondary_fname = '/'.join(target_fname.split('/')[:-(1 + parent_count)] +
                            secondary_fname.lstrip('/').split('/')[parent_count:])
        secondary_fname = "%s/%s" % (os.path.relpath(os.path.dirname(secondary_fname),
                                  os.path.dirname(source_fname)), os.path.basename(secondary_fname))
    target_fname = "%s/%s" % (os.path.relpath(os.path.dirname(target_fname),
                              os.path.dirname(source_fname)), os.path.basename(target_fname))
    print 'target: %s' % target_fname
    print 'secondary:', secondary_fname
    outfile = h5py.File(source_fname, 'a')
    if args.type == 'fends':
        outfile['/'].attrs['fendfilename'] = target_fname
    elif args.type == 'fragments':
        outfile['/'].attrs['fragfilename'] = target_fname
    if args.type == 'hic-data':
        if secondary_fname is not None:
            outfile['/'].attrs['fendfilename'] = secondary_fname
        outfile['/'].attrs['datafilename'] = target_fname
    if args.type == '5c-data':
        if secondary_fname is not None:
            outfile['/'].attrs['fragfilename'] = secondary_fname
        outfile['/'].attrs['datafilename'] = target_fname
    infile.close()
    outfile.close()
